# -*- coding: utf-8 -*-
"""
Modified by Group 04
"""

import random
import time
from halma_game.halma_player import HalmaPlayer
from halma_game.halma_model import HalmaModel


class HalmaPlayer01B(HalmaPlayer):
    def __init(self, nama):
        super().__init__(nama)


# -*- coding: utf-8 -*-
"""
Modified by Group 04
"""

import random
import time
from halma_game.halma_player import HalmaPlayer
from halma_game.halma_model import HalmaModel


class Node():
    parent = 0
    pos = 0

    # Node definition
    def __init__(self, parent=None, pos=None):
        self.parent = parent
        self.pos = pos

        # Heuristic value
        self.f = 0
        self.g = 0
        self.h = 0
    
    # Set heuristic value for node
    def set_heur(self, f, g, h):
        # f = g + h
        self.f = f
        self.g = g
        self.h = h

    # Check node position equality (boolean)
    def equal(self, other):
        return self.pos == other.pos


class HalmaPlayerHuman(HalmaPlayer):
    type = 'Human'

    def __init(self, nama):
        super().__init__(nama)
        

    # Calculate path from initial to final position using a-star search
    def calc_path(self, board, initial, final):
        def is_valid(pos):
            return (0 <= pos[0] < 10 and 0 <= pos[1] < 10)

        def is_board_piece(pos):
            return (board[pos[0]][pos[1]] != 0)

        print('Calculating path from {} to {}...'.format(initial, final))

        # Initial and final node
        initial_node = Node(None, initial)
        initial_node.set_heur(0,0,0)
        final_node = Node(None, final)
        final_node.set_heur(0,0,0)

        # List initialization (visit: to be visited)
        visit = []
        visit.append(initial_node)
        visited = []

        # Search loop (while there is node to be visited)
        while len(visit) > 0:

            # Select node with minimal f value of visit list
            current_node = visit[0]
            current_id = 0
            for index, node in enumerate(visit):
                if node.f < current_node.f:
                    current_node = node
                    current_id = index

            # Add visited current node to visited list
            visit.pop(current_id)
            visited.append(current_node)

            # If current node is final target
            if current_node.equal(final_node):
                path = []
                current = current_node
                
                # Backtracing parent position of node to initial position
                while current is not None:
                    path.append(current.pos)
                    current = current.parent
                
                # Return reversed path (from initial to final)
                return path[::-1]
            
            # Generate node children
            children = []
            move_possibilities = [(0,-1),(0,1),(-1,0),(1,0),(-1,-1),(-1,1),(1,-1),(1,1)]
            for move in move_possibilities:
                
                # Get next node position
                node_next_pos = (current_node.pos[0] + move[0], current_node.pos[1] + move[1])

                # Node position out of board index
                if not is_valid(node_next_pos):
                    continue
                
                # Next node position is board piece
                if is_board_piece(node_next_pos):
                    next2 = (node_next_pos[0] + move[0], node_next_pos[1] + move[1])

                    # Check if hop target is valid index for board
                    if is_valid(next2):

                        # Hop
                        if board[next2[0]][next2[1]] == 0:
                            node_next_pos = next2
                    
                        else:
                            continue
                    
                    else:
                        continue
                    
                # Create new node
                new_node = Node(current_node, node_next_pos)
                children.append(new_node)

            # Loop through node parent's children
            for child in children:

                # Child already on visited
                already = False
                for v in visited:
                    if child.equal(v):
                        already = True
                if already:
                    continue
                
                # Calculate A* parameter values (f,g,h)
                g = current_node.g + 1
                h = ((child.pos[0] - final_node.pos[0]) ** 2) + ((child.pos[1] - final_node.pos[1]) ** 2)
                child.set_heur(g + h, g, h)

                # Child already on visit
                already = False
                for v in visit:
                    if child.equal(v) and child.g > v.g:
                        already = True
                if already:
                    continue
                
                # Add child to visit list
                visit.append(child)
    


# -*- coding: utf-8 -*-
"""
Created on Sun Feb 23 07:23:50 2020

Congklak Board Game

@author: Mursito

Layout index:
-------
| 0 1 |
| 3 2 |
-------
"""

import time

# Global variable
N_KOTAK = 10
N_BIDAK_2 = 15
N_BIDAK_4 = 10

# Initial position (4P - 10 pieces)
ASAL_10_10_0=[(0,0),(0,1),(1,0),(0,2),(1,1),(2,0),(0,3),(1,2),(2,1),(3,0)]
ASAL_10_10_1=[(0,9),(0,8),(1,9),(0,7),(1,8),(2,9),(0,6),(1,7),(2,8),(3,9)]
ASAL_10_10_2=[(9,9),(9,8),(8,9),(9,7),(8,8),(7,9),(9,6),(8,7),(7,8),(6,9)]
ASAL_10_10_3=[(9,0),(9,1),(8,0),(9,2),(8,1),(7,0),(9,3),(8,2),(7,1),(6,0)]

# Initial position (2P - 15 pieces)
ASAL_10_15_0=[(0,0),(0,1),(1,0),(0,2),(1,1),(2,0),(0,3),(1,2),(2,1),(3,0),(0,4),(1,3),(2,2),(3,1),(4,0)]
ASAL_10_15_1=[(9,9),(9,8),(8,9),(9,7),(8,8),(7,9),(9,6),(8,7),(7,8),(6,9),(9,5),(8,6),(7,7),(6,8),(5,9)]

class HalmaModel:
    # Action
    A_GESER = 0
    A_LONCAT = 1
    A_BERHENTI = 2

    # Game action status
    S_OK = 0
    S_ILLEGAL = 1
    S_TIMEOUT = 2

    ARAH = [(-1,-1),(0,-1),(1,-1),(-1,0),(1,0),(-1,1),(0,1),(1,1)]

    JATAH_WAKTU = 10.0

    # Private variable
    __papan = []   
    __nkotak = 0
    __npemain = 0
    __pemain= []
    __giliran = 1
    __asal = []
    __tujuan = []
    __waktu = []
    __mulai = 0
    __menang = -1
    __teman = [2,3,0,1]

    # mulai main 2 pemain
    def awal(self, p1, p2, p3=None, p4=None):
        # 2-Player
        if (p3 == None) and (p4 == None):
            self.__npemain = 2
            self.__pemain = [p1, p2]
            self.__nbidak = N_BIDAK_2
            self.__asal = [ASAL_10_15_0, ASAL_10_15_1]
            self.__tujuan = [ASAL_10_15_1, ASAL_10_15_0]
        
        # 4-Player
        else:
            self.__npemain = 4
            self.__pemain = [p1, p2, p3, p4]
            self.__nbidak = N_BIDAK_4
            self.__asal = [ASAL_10_10_0, ASAL_10_10_1, ASAL_10_10_2, ASAL_10_10_3]
            self.__tujuan = [ASAL_10_10_2, ASAL_10_10_3, ASAL_10_10_0, ASAL_10_10_1]

        self.__nkotak = N_KOTAK
        self.__giliran = 0
        self.__papan = [[0]*self.__nkotak for i in range(self.__nkotak)]
        for i in range(self.__npemain):
            self.__pemain[i].setNomor(i+1)
            bp = (i + 1) * 100
            for j in range(self.__nbidak):
                x = self.__asal[i][j][0]
                y = self.__asal[i][j][1]
                self.__papan[x][y] = bp + (j + 1)
        self.__waktu = [0,0,0,0]


    # mengembalikan ukuran (N_KOTAK)
    def getUkuran(self):
        return self.__nkotak
    
    # mengembalikan jumlah pemain
    def getJumlahPemain(self):
        return self.__npemain

    # mengembalikan pemain
    def getPemain(self, ip):
        return self.__pemain[ip]

    # mengembalikan giliran
    def getGiliran(self):
        return self.__giliran

    # mengembalikan bidak di posisi x,y
    def getJumlahBidak(self):
        return self.__nbidak        

    # mengembalikan bidak di posisi x,y
    def getBidak(self, x, y):
        return self.__papan[x][y]        

    # mengembalikan semua bidak pemain tertentu
    # ini lama, jadi sebaiknya jangan dipanggil sering-sering
    def getPosisiBidak(self, p):
        bidak=[]
        bp = p+1
        for x in range(self.__nkotak):
            for y in range(self.__nkotak):
                bxy = self.__papan[x][y] // 100
                if (bxy == bp):
                    bidak.append((x,y))
        return bidak

    # mengembalikan dictionary nomor:posisi bidak pemain tertentu
    # ini lama, jadi sebaiknya jangan dipanggil sering-sering
    def getNomorPosisiBidak(self, p):
        bidak={}
        bp = p+1
        for x in range(self.__nkotak):
            for y in range(self.__nkotak):
                nxy = self.__papan[x][y]
                bxy = nxy // 100
                if (bxy == bp):
                    bidak[nxy] = (x,y)
        return bidak

    def getPapan(self):
        return self.__papan.copy()
    
    def getWaktu(self):
        return time.process_time()

    def getJatahWaktu(self, ip):
        return self.__waktu[ip]
    
    def setJatahWaktu(self, ip, delta):
        self.__waktu[ip] += delta

    def getSisaWaktu(self):
        ip = self.__giliran
        return self.__waktu[ip] - (time.process_time()-self.__mulai)
                            
    # return true jika x,y masih dalam papan
    def dalamPapan(self, x2, y2):
        if (x2 < 0) or (x2 >= self.__nkotak):
            return False
        if (y2 < 0) or (y2 >= self.__nkotak):
            return False        
        return True

    # return true jika x,y dalam area tujuan
    def dalamTujuan(self, ip, x, y):
        for xy in self.__tujuan[ip]:
            if (xy[0] == x) and (xy[1]==y):
                return True
        return False
    
    # return true jika boleh geser dr x1,y1 ke x2,y2
    def bolehGeser(self, ip, x1, y1, x2, y2):
        if not self.dalamPapan(x2, y2):
            return False
        if (self.__papan[x2][y2] != 0):
            return False
        for a in self.ARAH:
            x21 = x1 + a[0]
            y21 = y1 + a[1]
            if (x21==x2) and (y21==y2):
                return True
        return False

    # return true jika boleh loncat dr x1,y1 ke x2,y2
    def bolehLoncat(self, ip, x1, y1, x2, y2):
        if not self.dalamPapan(x2, y2):
            return False
        if (self.__papan[x2][y2] != 0):
            return False
        for a in self.ARAH:
            x21 = x1 + a[0] + a[0]
            y21 = y1 + a[1] + a[1]
            if (x21==x2) and (y21==y2):
                return True
        return False

    # mulai main, akan dicatat waktunya
    # setelah itu pemain bisa tanya dengan getSisaWaktu
    def mainMulai(self):
        ip = self.__giliran
        self.__mulai = time.process_time()

    # jalan satu geseran
    # return false kalau tak boleh
    def mainGeser(self, x1, y1, x2, y2):
        bnum = self.__papan[x1][y1]
        ip = (bnum // 100) - 1

        if (ip != self.__giliran):
            return self.S_ILLEGAL
        if not self.bolehGeser(ip, x1, y1, x2, y2):
            return self.S_ILLEGAL            
        self.__papan[x2][y2] = self.__papan[x1][y1]
        self.__papan[x1][y1] = 0
        return self.S_OK

    # jalan satu loncatan
    # return false kalau tak boleh
    def mainLoncat(self, x1, y1, x3, y3):
        bnum = self.__papan[x1][y1]
        ip = (bnum // 100) - 1
        if (ip != self.__giliran):
            return self.S_ILLEGAL
        if not self.bolehLoncat(ip, x1, y1, x3, y3):
            return self.S_ILLEGAL
        self.__papan[x3][y3] = self.__papan[x1][y1]
        self.__papan[x1][y1] = 0
        return self.S_OK


    # periksa apakah sudah berakhir
    # return True jika sudah berakhir 
    def akhir(self):
        bp = self.__giliran+1
        for xy in self.__tujuan[self.__giliran]:
            bxy = self.__papan[xy[0]][xy[1]] // 100
            if (bxy != bp):
                return False
        return True
    
    # ganti pemain berikutnya, sambil periksa waktu
    # return True jika pemain lama masih punya jatah waktu
    def ganti(self, selesai):
        self.__waktu[self.__giliran] += self.JATAH_WAKTU - (selesai - self.__mulai)
        # print('time stack: ', self.__waktu[self.__giliran])
        if self.__waktu[self.__giliran] < 0:
            return self.S_TIMEOUT
        self.__giliran = (self.__giliran + 1) % self.__npemain
        return self.S_OK

    # periksa apakah regu sudah berakhir
    # return True jika sudah berakhir 
    def akhirBeregu(self):
        bp = self.__giliran+1
        for xy in self.__tujuan[self.__giliran]:
            bxy = self.__papan[xy[0]][xy[1]] // 100
            if (bxy != bp):
                return False

        teman = self.__teman[self.__giliran]
        bp = teman+1
        for xy in self.__tujuan[teman]:
            bxy = self.__papan[xy[0]][xy[1]] // 100
            if (bxy != bp):
                return False
                
        return True

    '''
        Added function
    '''

    def is_valid(self, pos):
        return ((0 <= pos[0] < 10) and (0 <= pos[1] < 10))
    

    def is_empty(self, pos):
        return self.__papan[pos[0]][pos[1]] == 0


    def is_board_piece(self, pos):
        return self.__papan[pos[0]][pos[1]] // 100


    # Get all board pieces position of player_id (list of position)
    def get_board_pieces(self, player_id):
        pieces = []

        # Board traversal
        for i in range(0,10):
            for j in range(0,10):
                if self.is_board_piece((i,j)) == player_id:
                    pieces.append((i,j))
        return pieces
    

    def get_loncat_multi(self, pos, player_id):
    
        # Helper
        def get_loncat(pos):
            loncat = []

            # Explore all move direction
            for move in self.ARAH:
                x1 = pos[0] + move[0]
                y1 = pos[1] + move[1]
                x2 = pos[0] + (2 * move[0])
                y2 = pos[1] + (2 * move[1])
                
                # Check if can hop and add to list
                if self.is_valid((x2,y2)):
                    if self.is_board_piece((x1,y1)) and self.is_empty((x2,y2)):
                        loncat.append((x2,y2))
            
            return loncat

        # Main
        hopped = []
        path = []
        path_queue = [[pos]]

        while path_queue != []:
            current_path = path_queue.pop(0)
            current_node = current_path[-1]
            hopped.append(current_node)

            hopping_child = get_loncat(current_node)
            h_num = 0
            for h_child in hopping_child:
                if h_child not in hopped:
                    hopped.append(h_child)
                    path_queue.append(current_path + [h_child])
                    h_num += 1

            if h_num == 0:
                path.append(current_path)

        # return [[(x2,y2),(x3,y3)],[(x2,y2),(x3,y3),(x4,y4)],...]
        return path


    # Get all hop possibilities for each board piece in node object
    def get_all_loncat(self, player_id):
        loncat_all = []
        pieces = self.get_board_pieces(player_id)

        # Explore all pieces
        for piece in pieces:
            # Get hop list for piece
            hop_list = []
            for h in self.get_loncat_multi(piece, player_id):
                for n in range(1,len(h)):
                    h_buf = h[1:n+1]
                    if h_buf not in hop_list:
                        hop_list.append(h_buf)
            
            if hop_list != [[]]:
                loncat_all.append([piece,hop_list])
        
        return loncat_all
    

    # Get all geser possibilities for each board piece in node object
    def get_all_geser(self, player_id):
        geser = []
        pieces = self.get_board_pieces(player_id)
        for piece in pieces:
            geser_buf = []
            for move in self.ARAH:
                x1 = piece[0] + move[0]
                y1 = piece[1] + move[1]
                
                if self.is_valid((x1,y1)) and self.is_empty((x1,y1)):
                    geser_buf.append((x1,y1))

            geser.append([piece,geser_buf])
        
        return geser


    # Get all moves (return list hop,geser)
    def get_all_moves(self, player_id):
        moves = []

        # Hop
        moves_hop = self.get_all_loncat(player_id)
        for piece_hop in range(len(moves_hop)):
            for move in moves_hop[piece_hop][1]:
                moves.append([moves_hop[piece_hop][0]] + move + ['1'])
        
        # Geser
        moves_geser = self.get_all_geser(player_id)
        for piece_geser in range(len(moves_geser)):
            for move in moves_geser[piece_geser][1]:
                moves.append([moves_geser[piece_geser][0]] + [move] + ['0'])
        
        # print('get_all_moves',moves)
        return moves


    # Calculate path from initial to final position using a-star search
    def calc_path_hop(self, initial, final):
        def is_valid(pos):
            return (0 <= pos[0] < 10 and 0 <= pos[1] < 10)

        def is_board_piece(pos):
            return (self.__papan[pos[0]][pos[1]] != 0)

        # print('Calculating path from {} to {}...'.format(initial, final))

        # Initial and final node
        initial_node = Node(None, initial)
        initial_node.set_heur(0,0,0)
        final_node = Node(None, final)
        final_node.set_heur(0,0,0)

        # List initialization (visit: to be visited)
        visit = []
        visit.append(initial_node)
        visited = []

        # Search loop (while there is node to be visited)
        while len(visit) > 0:

            # Select node with minimal f value of visit list
            current_node = visit[0]
            current_id = 0
            for index, node in enumerate(visit):
                if node.f < current_node.f:
                    current_node = node
                    current_id = index

            # Add visited current node to visited list
            visit.pop(current_id)
            visited.append(current_node)

            # If current node is final target
            if current_node.equal(final_node):
                path = []
                current = current_node
                
                # Backtracing parent position of node to initial position
                while current is not None:
                    path.append(current.pos)
                    current = current.parent
                
                # Return reversed path (from initial to final)
                return path[::-1]
            
            # Generate node children
            children = []
            move_possibilities = [(0,-1),(0,1),(-1,0),(1,0),(-1,-1),(-1,1),(1,-1),(1,1)]
            for move in move_possibilities:
                
                # Get next node position
                node_next_pos = (current_node.pos[0] + move[0], current_node.pos[1] + move[1])

                # Node position out of board index
                if not is_valid(node_next_pos):
                    continue
                
                # Next node position is board piece
                if is_board_piece(node_next_pos):
                    next2 = (node_next_pos[0] + move[0], node_next_pos[1] + move[1])

                    # Check if hop target is valid index for board
                    if is_valid(next2):

                        # Hop
                        if self.is_empty(next2):
                            node_next_pos = next2
                            
                            # Create new node
                            new_node = Node(current_node, node_next_pos)
                            children.append(new_node)
                    
                        else:
                            continue
                    
                    else:
                        continue
                    
                

            # Loop through node parent's children
            for child in children:

                # Child already on visited
                already = False
                for v in visited:
                    if child.equal(v):
                        already = True
                if already:
                    continue
                
                # Calculate A* parameter values (f,g,h)
                g = current_node.g + 1
                h = ((child.pos[0] - final_node.pos[0]) ** 2) + ((child.pos[1] - final_node.pos[1]) ** 2)
                child.set_heur(g + h, g, h)

                # Child already on visit
                already = False
                for v in visit:
                    if child.equal(v) and child.g > v.g:
                        already = True
                if already:
                    continue
                
                # Add child to visit list
                visit.append(child)

# Supporting Class
class Node():
    parent = 0
    pos = 0

    # Node definition
    def __init__(self, parent=None, pos=None):
        self.parent = parent
        self.pos = pos

        # Heuristic value
        self.f = 0
        self.g = 0
        self.h = 0
    
    # Set heuristic value for node
    def set_heur(self, f, g, h):
        # f = g + h
        self.f = f
        self.g = g
        self.h = h

    # Check node position equality (boolean)
    def equal(self, other):
        return self.pos == other.pos


# -*- coding: utf-8 -*-
"""
Created on Sun Feb 23 08:04:48 2020

@author: Mursito
"""

import random
import time
from halma_game.halma_model import HalmaModel

class HalmaPlayer:
    nama = 'Pemain'
    deskripsi = 'Random Strategy'
    type = 'AI'
    nomor = 1    
    index = 0
    papan = []
    teman = None
    
    def __init__(self, nama):
        self.nama = nama
        
    def setNomor(self, nomor):
        self.nomor = nomor
        self.index = nomor-1
    
    def setTeman(self, p):
        self.teman = p
    
    def getTeman(self):
        return self.teman
    
    def getType(self):
        return self.type

    # mengembalikan semua kemungkikan main (geser / loncat) bidak di (x1, y1)
    def bisaMain(self, model, papan, x1, y1):
        geser = []
        loncat = []
        ip = self.index
        dTujuan = model.dalamTujuan(ip, x1, y1)
        for a in model.ARAH:
            x2 = x1 + a[0]
            y2 = y1 + a[1]
            #print((x2, y2), end="")
            if model.dalamPapan(x2, y2):
                if (papan[x2][y2] == 0):
                    if not dTujuan or model.dalamTujuan(ip, x2, y2):
                        geser.append((x2,y2))
                else:
                    x3 = x2 + a[0]
                    y3 = y2 + a[1]
                    #print((x3, y3), end="")
                    if model.dalamPapan(x3, y3):
                        if (papan[x3][y3] == 0):
                            if not dTujuan or model.dalamTujuan(ip, x3, y3):
                                loncat.append((x3,y3))
        return geser, loncat
       
    # Pemain beraksi
    # return [(x2,y2)], (x1,y1), aksi
    # aksi = A_GESER, A_LONCAT, atau A_BERHENTI
    # (x1, y1) = posisi bidak awal
    # [(x2, y2)] = posisi tujuan (array, isi 1 kalau geser, isi banyak kalau loncat)
    def main(self, model):
        random.seed(time.process_time_ns())
        papan = model.getPapan()
        b0 = model.getPosisiBidak(self.index)
        random.shuffle(b0)

        k = 0
        # Dummy algorithms for delay
        for i in range(10000000):
            k += i

        for b in b0:
            g,l = self.bisaMain(model, papan, b[0], b[1])
            
            # Randomize move
            random.shuffle(g)
            random.shuffle(l)

            if g != [] :
                #print('return g',[g[0]],b,model.A_GESER)
                return [g[0]], b, model.A_GESER
            if l != [] :
                #print('return l',[l[0]],b,model.A_LONCAT)
                return [l[0]], b, model.A_LONCAT
        return None, None, model.A_BERHENTI


# Player AI script import
from halma_game.halma_model import HalmaModel
from halma_game.halma_player import HalmaPlayer
from halma_game.halma_player_01_A import HalmaPlayer01A
from halma_game.halma_player_01_B import HalmaPlayer01B
from halma_game.halma_player_02_A import HalmaPlayer02A
from halma_game.halma_player_02_B import HalmaPlayer02B
from halma_game.halma_player_03_A import HalmaPlayer03A
from halma_game.halma_player_03_B import HalmaPlayer03B
from halma_game.halma_player_04_A import HalmaPlayer04A
from halma_game.halma_player_04_B import HalmaPlayer04B
from halma_game.halma_player_human import HalmaPlayerHuman


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="halma-game",
    package=['halma-game'],
    version="0.3.5",
    license='MIT',
    author="Rasis Syauqi Buldan",
    author_email="rasisbuldan@gmail.com",
    description="Halma game created with python (pygame)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rasisbuldan/halma_game",
    download_url="https://github.com/rasisbuldan/halma_game/archive/v0.3.5.tar.gz",
    packages=setuptools.find_packages(),
    install_requires=[
        'pygame',
      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    include_package_data=True
)


# -*- coding: utf-8 -*-
"""
Gen 3 - Team 04

Recursive minimax with alpha-beta pruning
Evaluation (weighted):
    - [A/B] Distance to endpoint (0,0), (9,9), (0,9), (9,0)
    - [A/B] Middle lane
    - [A] Hop possibilities and count for ally
    - [B] Max (Furthest hop or delta distance)
    - Hop possibilities and count for enemy
    - Potential delta

To do:
    - add move history constraint
    - mid lane control
    - a* search piece-blockade error
"""

import random
import time
from halma_game.halma_player import HalmaPlayer


class HalmaStateNode:
    move_dir = (0, [(1,-1),(1,0),(0,-1),(1,1),(-1,-1),(0,1),(-1,0),(-1,1)],     # Player 1
                   [(-1,-1),(0,-1),(-1,0),(-1,1),(1,-1),(0,1),(1,0),(1,1)],     # Player 2
                   [(-1,1),(-1,0),(0,1),(-1,-1),(1,1),(0,-1),(1,0),(1,-1)],     # Player 3
                   [(1,1),(0,1),(1,0),(-1,1),(1,-1),(-1,0),(0,-1),(-1,-1)])     # Player 4
    val = 0

    def __init__(self, turn, parent=None, current=None, move=None):
        # Store attributes value
        self.parent = parent
        self.current = [row[:] for row in current]      # 2-D array deepcopy
        self.move = move
        self.turn = turn
        # print(move)

        # Run move [(x0,y0),(x1,y1),1]
        # print(*self.current, sep='\n')
        # print('move: ', move)
        if move != None:
            self.current[move[-2][0]][move[-2][1]] = self.current[move[0][0]][move[0][1]]
            self.current[move[0][0]][move[0][1]] = 0
        # print(*self.current, sep='\n')

    # ---- Selector ----
    def get_move(self):
        return self.move

    def get_current(self):
        return [row[:] for row in self.current]

    def get_parent(self):
        return self.parent
    
    def get_val(self):
        return self.val
    
    def get_turn(self):
        return self.turn
    
    def is_valid(self, pos):
        return ((0 <= pos[0] < 10) and (0 <= pos[1] < 10))
    
    def is_empty(self, pos):
        return self.current[pos[0]][pos[1]] == 0

    def is_board_piece(self, pos):
        # print('[is_board_piece]',self.get_turn(),'->',pos,'->',self.current[pos[0]][pos[1]] // 100)
        return self.current[pos[0]][pos[1]] // 100
    
    # Return initial,final
    def get_dist_reference(self, player_id):
        if player_id == 1:
            return (0,0),(9,9)
        elif player_id == 2:
            return (0,9),(9,0)
        elif player_id == 3:
            return (9,9),(0,0)
        elif player_id == 4:
            return (9,0),(0,9)
    
    # Pieces with pos in triangle zone
    def in_zone(self, pos, player_id):
        # Player 1 (Top left)
        if player_id == 1:
            return pos[0] in range(0,4) and pos[1] in range(0, 4 - pos[0])

        # Player 2 (Top right)
        elif player_id == 2:
            return pos[0] in range(0,4) and pos[1] in range(6 + pos[0], 10)

        # Player 3 (Bottom right)
        elif player_id == 3:
            return pos[0] in range(6,10) and pos[1] in range(10 - (pos[0] - 5), 10)

        # Player 4 (Bottom left)
        elif player_id == 4:
            return pos[0] in range(6,10) and pos[1] in range(0, pos[0] - 5)

    # Pieces with pos in target initial zone
    def in_target_zone(self, pos, player_id):
        return self.in_zone(pos, 1 + ((player_id + 1) % 4))

    def count_in_base(self, player_id):
        pieces = self.get_board_pieces(player_id)

        k = 0
        for piece in pieces:
            if self.in_zone(piece, player_id):
                k += 1
        
        return k

    # Count board pieces in target zone
    def count_in_target(self, player_id):
        pieces = self.get_board_pieces(player_id)

        k = 0
        for piece in pieces:
            if self.in_target_zone(piece, player_id):
                k += 1
        
        return k
    

    def get_empty_target(self, player_id):
        pieces = self.get_board_pieces(player_id)
        empty_target = []

        for i in range (0,10):
            for j in range(0,10):
                if self.in_target_zone((i,j),player_id) and (self.is_empty((i,j)) or self.is_board_piece((i,j)) != player_id):
                    empty_target.append((i,j))
        
        return empty_target

    def get_lost_piece(self, player_id):
        pieces = self.get_board_pieces(player_id)
        lost_piece = []

        for piece in pieces:
            if not self.in_target_zone(piece, player_id):
                lost_piece.append(piece)
        
        return lost_piece


    # Game finished check function
    def game_finished(self):
        return self.count_in_target(self.get_turn()) == 10

    # Get all board pieces position of player_id (list of position)
    def get_board_pieces(self, player_id):
        pieces = []

        # Board traversal
        for i in range(0,10):
            for j in range(0,10):
                if self.is_board_piece((i,j)) == player_id:
                    pieces.append((i,j))
        return pieces


    def get_loncat_multi(self, pos, player_id):

        # Helper
        def get_loncat(pos):
            loncat = []

            # Explore all move direction
            for move in self.move_dir[player_id]:
                x1 = pos[0] + move[0]
                y1 = pos[1] + move[1]
                x2 = pos[0] + (2 * move[0])
                y2 = pos[1] + (2 * move[1])
                
                # Check if can hop and add to list
                if self.is_valid((x2,y2)):
                    if self.is_board_piece((x1,y1)) and self.is_empty((x2,y2)):
                        loncat.append((x2,y2))
            
            return loncat

        # Main
        hopped = []
        path = []
        path_queue = [[pos]]

        while path_queue != []:
            current_path = path_queue.pop(0)
            current_node = current_path[-1]
            hopped.append(current_node)

            hopping_child = get_loncat(current_node)
            h_num = 0
            for h_child in hopping_child:
                if h_child not in hopped:
                    hopped.append(h_child)
                    path_queue.append(current_path + [h_child])
                    h_num += 1

            if h_num == 0:
                path.append(current_path)

        # return [[(x2,y2),(x3,y3)],[(x2,y2),(x3,y3),(x4,y4)],...]
        return path


    # Get all hop possibilities for each board piece in node object
    def get_all_loncat(self, player_id):
        loncat_all = []
        pieces = self.get_board_pieces(player_id)

        # Explore all pieces
        for piece in pieces:
            # Get hop list for piece
            hop_list = []
            for h in self.get_loncat_multi(piece, player_id):
                for n in range(1,len(h)):
                    h_buf = h[1:n+1]
                    if h_buf not in hop_list:
                        hop_list.append(h_buf)
            
            if hop_list != [[]]:
                loncat_all.append([piece,hop_list])
        
        return loncat_all

    # Get furthest hop
    # hop = [[(0,0),(1,1),(1,2),(1,3),(1,4)],
	#        [(0,0),(1,1),(2,1),(2,2)]]
    def get_furthest_loncat(self, player_id):
        max_hop_val = 0
        max_hop = []

        initial, target = self.get_dist_reference(player_id)

        moves_hop = self.get_all_loncat(player_id)
        for piece_hop in range(len(moves_hop)):
            for move in moves_hop[piece_hop][1]:
                val = self.calc_dist_normalized(move[-1],initial,target)
                
                if val > max_hop_val:
                    max_hop = [moves_hop[piece_hop][0]] + move
                    max_hop_val = val
        
        # print('~~ [max hop]',player_id,'->',max_hop)
        return max_hop

    
    # Get hop count (all player)
    def get_loncat_count(self, player_id):
        moves_hop = self.get_all_loncat(player_id)
        hop_count = 0

        for piece_hop in range(len(moves_hop)):
            hop_count += len(moves_hop[piece_hop][1])

        return hop_count
    

    # Get all geser possibilities for each board piece in node object
    def get_all_geser(self, player_id):
        geser = []
        pieces = self.get_board_pieces(player_id)
        for piece in pieces:
            geser_buf = []
            for move in self.move_dir[player_id]:
                x1 = piece[0] + move[0]
                y1 = piece[1] + move[1]
                
                if self.is_valid((x1,y1)) and self.is_empty((x1,y1)):
                    geser_buf.append((x1,y1))

            geser.append([piece,geser_buf])
        
        return geser
    

    # Moves constraint
    # Previous : in move history
    def constraint(self, move):
        return True


    # Get all moves (return list hop,geser)
    def get_all_moves(self, player_id):
        moves = []

        # Hop
        moves_hop = self.get_all_loncat(player_id)
        for piece_hop in range(len(moves_hop)):
            for move in moves_hop[piece_hop][1]:
                if self.constraint(move):
                    moves.append([moves_hop[piece_hop][0]] + move + ['1'])
        
        # Geser
        moves_geser = self.get_all_geser(player_id)
        for piece_geser in range(len(moves_geser)):
            for move in moves_geser[piece_geser][1]:
                if self.constraint(move):
                    moves.append([moves_geser[piece_geser][0]] + [move] + ['0'])
        
        return moves


    # Calculate chebyshev distance (omnidirectional) from initial to final (tuple)
    def calc_chebyshev(self, initial, final):
        return max(abs(final[0] - initial[0]), abs(final[1] - initial[1]))


    # Calculate euclidean distance
    def calc_dist(self, initial, final):
        return (abs(final[0] - initial[0]) ** 2) + (abs(final[1] - initial[1]) ** 2)


    # Normalize evaluation parameter (dist=0 -> val=1)
    def calc_dist_normalized(self, pos, initial, final):
        max_val = (abs(final[0] - initial[0]) ** 2) + (abs(final[1] - initial[1]) ** 2)
        val = (abs(final[0] - pos[0]) ** 2) + (abs(final[1] - pos[1]) ** 2)
        return 1 - (val / max_val)
        
    
    # Calculate current state evaluation function (ally + enemy -)
    def calc_evaluation(self, weight):
        root_id = 1 + (self.turn % 4)
        eval_value = 0

        for i in range(0,4):
            turn_id = 1 + ((self.turn + i) % 4)
            pieces = self.get_board_pieces(turn_id)
            
            if i % 2 == 0:
                dist_val = 0
                for piece in pieces:
                    dist_val += self.calc_dist_normalized(piece, *self.get_dist_reference(turn_id))
                
                # Pieces total distance to target zone
                eval_value += (weight['dist-target'] * (dist_val / 10))
                # Pieces count in target zone
                eval_value += (weight['count-target'] * (self.count_in_target(turn_id) / 10))
                
                # Pieces count in base zone
                #eval_value -= (weight['count-base'] * (self.count_in_base(turn_id) / 10))

                # Furthest hop distance to target zone
                #eval_value += (weight['furthest-loncat'] * self.calc_dist_normalized(self.get_furthest_loncat(turn_id)[-1], 
                                                                #*self.get_dist_reference(turn_id)))\
                
                # Hop count
                #eval_value += (weight['count-loncat'] * self.get_loncat_count(turn_id) / 50)
        
        # Save evaluation value to val attribute
        # print('~~ [eval_value]',self.move,'->',eval_value)
        self.val = eval_value

    
class Node():
    parent = 0
    pos = 0

    # Node definition
    def __init__(self, parent=None, pos=None):
        self.parent = parent
        self.pos = pos

        # Heuristic value
        self.f = 0
        self.g = 0
        self.h = 0
    
    # Set heuristic value for node
    def set_heur(self, f, g, h):
        # f = g + h
        self.f = f
        self.g = g
        self.h = h

    # Check node position equality (boolean)
    def equal(self, other):
        return self.pos == other.pos



class HalmaPlayer04B(HalmaPlayer):
    # Class Attributes
    last_move = [] # Move history (limited to n-move)
    ply = 1
    late_treshold = 0
    time_start = 0
    turn_count = 0

    def __init(self, nama):
        super().__init__(nama)
        self.moves = []
        self.turn_count = 0

    # Special case - late game
    # Calculate path using a-star search
    def calc_path(self, board, initial, final):
        def is_valid(pos):
            return (0 <= pos[0] < 10 and 0 <= pos[1] < 10)

        def is_board_piece(pos):
            return (board[pos[0]][pos[1]] != 0)

        # print('Calculating path...')

        # Initial and final node
        initial_node = Node(None, initial)
        initial_node.set_heur(0,0,0)
        final_node = Node(None, final)
        final_node.set_heur(0,0,0)

        # List initialization (visit: to be visited)
        visit = []
        visit.append(initial_node)
        visited = []

        # Search loop (while there is node to be visited)
        while len(visit) > 0:

            # Select node with minimal f value of visit list
            current_node = visit[0]
            current_id = 0
            for index, node in enumerate(visit):
                if node.f < current_node.f:
                    current_node = node
                    current_id = index
            # print('visiting ',current_node.pos)

            # Add visited current node to visited list
            visit.pop(current_id)
            visited.append(current_node)

            # If current node is final target
            if current_node.equal(final_node):
                # print('[final node]')
                path = []
                current = current_node
                
                # Backtracing parent position of node to initial position
                while current is not None:
                    path.append(current.pos)
                    # print('[final path] ',path)
                    current = current.parent
                
                # Return reversed path (from initial to final)
                return path[::-1]
            
            # Generate node children
            children = []
            move_possibilities = [(0,-1),(0,1),(-1,0),(1,0),(-1,-1),(-1,1),(1,-1),(1,1)]
            for move in move_possibilities:
                
                # Get next node position
                node_next_pos = (current_node.pos[0] + move[0], current_node.pos[1] + move[1])

                # Node position out of board index
                if not is_valid(node_next_pos):
                    continue
                
                # Next node position is board piece
                if is_board_piece(node_next_pos):
                    next2 = (node_next_pos[0] + move[0], node_next_pos[1] + move[1])

                    # Check if hop target is valid index for board
                    if is_valid(next2):

                        # Hop
                        if board[next2[0]][next2[1]] == 0:
                            node_next_pos = next2
                    
                        else:
                            continue
                    
                    else:
                        continue
                    
                # Create new node
                new_node = Node(current_node, node_next_pos)
                children.append(new_node)

            # Loop through node parent's children
            for child in children:

                # Child already on visited
                already = False
                for v in visited:
                    if child.equal(v):
                        already = True
                if already:
                    continue
                
                # Calculate A* parameter values (f,g,h)
                g = current_node.g + 1
                h = ((child.pos[0] - final_node.pos[0]) ** 2) + ((child.pos[1] - final_node.pos[1]) ** 2)
                child.set_heur(g + h, g, h)

                # Child already on visit
                already = False
                for v in visit:
                    if child.equal(v) and child.g > v.g:
                        already = True
                if already:
                    continue
                
                # Add child to visit list
                # print('appending ',child.pos)
                visit.append(child)


    def minimax_node(self, turn, node, depth, alpha, beta, weight, maximizingPlayer):
        # Termination (ply reached, game finished, or time limit)
        if depth == self.ply or node.game_finished() or (time.process_time() - self.time_start) > self.time_limit:
            node.calc_evaluation(weight)
            # print('~~ [node move]',node.get_move(),' ==> ',node.get_val())
            return node
        
        # Our turn (depth 2n - 1)
        if maximizingPlayer:
            max_eval = -9999999

            # Get max move
            max_moves = node.get_all_moves(node.get_turn())

            # Moves traversal
            for move in max_moves:
                h = HalmaStateNode(turn, node, node.get_current(), move)
                node_eval = self.minimax_node(1 + (turn % 4), h, depth + 1, alpha, beta, weight, False)
                
                # Get max value between alpha and node evaluation value
                if node_eval.get_val() > max_eval:
                    max_eval = node_eval.get_val()
                    max_node = node_eval
                
                alpha = max(node_eval.get_val(), alpha)
                
                # Pruning
                if beta <= alpha:
                    # print('prune!')
                    break
            
            return max_node

        # Opponent turn (depth 2n)
        else:
            min_eval = 9999999

            # Get min move
            min_moves = node.get_all_moves(node.get_turn())

            # Moves traversal
            for move in min_moves:
                h = HalmaStateNode(turn, node, node.get_current(), move)
                node_eval = self.minimax_node(1 + (turn % 4), h, depth + 1, alpha, beta, weight, True)

                # Get min value between beta and node evaluation value
                if node_eval.get_val() < min_eval:
                    min_eval = node_eval.get_val()
                    min_node = node_eval

                beta = min(node_eval.get_val(), beta)
                
                # Pruning
                if beta <= alpha:
                    # print('prune!')
                    break

            return min_node


    # Called API for model
    # Return format :
    #    GESER:     [(x1,y1)], (x0,y0), 0
    #    LONCAT:    [(x1,y1),(x2,y2),...], (x0,y0), 1
    #    BERHENTI:  None, None, 2
    def main(self, model):
        # AI parameter
        self.ply = 2
        self.time_limit = 3
        self.late_treshold = 9
        weight = {
            'dist-target': 0.5,
            'count-target': 0.3,
            'count-base': 0.15,
            'count-loncat': 0.05,
            'furthest-loncat': 0.2
        }

        # Algorithm parameter
        alpha = -9999999
        beta = 9999999
        board = [row[:] for row in model.getPapan()]    # 2-D array deepcopy

        # Time and move count tracking
        self.time_start = time.process_time()
        self.turn_count += 1

        root_node = HalmaStateNode(self.nomor, None, board, None)
        #print('~~ [target]',root_node.get_turn(),'-->',root_node.get_dist_reference(root_node.get_turn()))

        if not root_node.game_finished():
            # Late game
            if root_node.count_in_target(self.nomor) >= 9:

                # print('!! [late game]', self.nomor)
                # print(*root_node.get_current(), sep='\n')
                initial = root_node.get_lost_piece(self.nomor)[0]
                target = root_node.get_empty_target(self.nomor)[0]
                board = root_node.get_current()

                final = self.calc_path(board, initial, target)
                # print(final)

                if (abs(final[1][0] - initial[0]) ** 2 + abs(final[1][1] - initial[1]) ** 2) > 2:
                    return [final[1]], initial, 1
                else:
                    return [final[1]], initial, 0

            
            # Early-mid game
            else:
                # Binary tree exploration
                node_choice = self.minimax_node(self.nomor, root_node, 0, alpha, beta, weight, True)
                val = node_choice.get_val()
                # print('!! [node choice]',node_choice.get_move(),'->',node_choice.get_val())

                # Backtrace to selected first child node of root_node
                while node_choice.get_parent().get_parent() != None:
                    # print('!! [node parent]',node_choice.get_parent().get_move())
                    node_choice = node_choice.get_parent()
                
                # Extract move information
                # print('!! [final node choice]',node_choice.get_move(),'->',val)
                move_choice = node_choice.get_move()
            
                initial = move_choice[0]
                act_num = int(move_choice[-1])

                # Geser
                if act_num == 0:
                    final = [move_choice[1]]
                    action = model.A_GESER
                # Loncat
                elif act_num == 1:
                    final = move_choice[1:-1]
                    action = model.A_LONCAT
                
                # print('!! [return]',final,',',initial,',',action)
                return final, initial, action
        
        # Game finished
        else:
            # print('!! [return] no move')
            return None,None,model.A_BERHENTI


''' 
Halma GUI (Using halma_model script)

To do improvement:
    - Continuous timer (while AI computing moves)
    - Human player
    - Debug/Strategy mode

To do:
    - Animate move
    - 4 player color picker (multiple click)
    - Pause button (stop timer tick)
    - Button create from text
    - Dictionary asset import
    - Move history line
    - Decompose script to module halma_display
    - Remove 8x8 support?
    - Verbose / quiet parameter
    - Reset function redefinition
    - Offset font size with Font.size()
    - Analytics (avg time, move arrow)
    - Resolution (scaling?)
    - PyPi package

To do test:
    - Font size

'''


# Module import
import pygame
import pygame.gfxdraw
import time
import math
import os
import sys


# Pygame initialization
pygame.init()
screen = pygame.display.set_mode((1280,720))    # Resolution set
pygame.display.set_caption('Halma (Beta) v0.3.5')
pygame.display.set_icon(pygame.image.load('assets/icon.png'))


# Color Definition
colors = {
    'BLACK':        (0,0,0),
    'WHITE':        (255,255,255),
    'BG':           (236,243,244),
    'BG_DARK':      (10,31,34),
    'TEXT':         (40,76,81),
    'TEXT_DARK':    (236,243,244),
    'BLUE':         (69,165,229),
    'CYAN':         (116,254,220),
    'GREEN':        (87,184,77),
    'ORANGE':       (243,156,18),
    'PINK':         (253,153,204),
    'PURPLE':       (155,89,182),
    'RED':          (240,111,82),
    'YELLOW':       (241,196,15)
}
colors_enum = list(enumerate(colors))


# Assets Import
title = {
    'light'             : pygame.image.load('assets/title.png'),
    'dark'              : pygame.image.load('assets/title_dark.png')
}
info_text = {
    'turn'              : pygame.image.load('assets/info/info_turn.png'),
    'score'             : pygame.image.load('assets/info/info_score.png'),
    'time-left'         : pygame.image.load('assets/info/info_timeleft.png'),
    'last-move'         : pygame.image.load('assets/info/info_lastmove.png')
}
info_player = {
    'p1'                : pygame.image.load('assets/info/info_p1.png'),
    'p2'                : pygame.image.load('assets/info/info_p2.png')
}
button = {                   
    'start'             : pygame.image.load('assets/button/button_start.png'),
    'start-active'      : pygame.image.load('assets/button/button_start_active.png'),
    'reset'             : pygame.image.load('assets/button/button_reset.png'),
    'reset-active'      : pygame.image.load('assets/button/button_reset_active.png'),
    'pause'             : pygame.image.load('assets/button/button_pause.png'),
    'pause-active'      : pygame.image.load('assets/button/button_pause_active.png'),
    '8x8'               : pygame.image.load('assets/button/button_8x8.png'),
    '8x8-active'        : pygame.image.load('assets/button/button_8x8_active.png'),
    '10x10'             : pygame.image.load('assets/button/button_10x10.png'),
    '10x10-active'      : pygame.image.load('assets/button/button_10x10_active.png'),
    '2p'                : pygame.image.load('assets/button/button_2p.png'),
    '2p-active'         : pygame.image.load('assets/button/button_2p_active.png'),
    '4p'                : pygame.image.load('assets/button/button_4p.png'),
    '4p-active'         : pygame.image.load('assets/button/button_4p_active.png'),
    'dark'              : pygame.image.load('assets/button/button_dark.png'),
    'dark-active'       : pygame.image.load('assets/button/button_dark_active.png')
}
button_player = {
    '01'                : pygame.image.load('assets/button/button_01.png'),
    '01-active'         : pygame.image.load('assets/button/button_01_active.png'),
    '02'                : pygame.image.load('assets/button/button_02.png'),
    '02-active'         : pygame.image.load('assets/button/button_02_active.png'),
    '03'                : pygame.image.load('assets/button/button_03.png'),
    '03-active'         : pygame.image.load('assets/button/button_03_active.png'),
    '04'                : pygame.image.load('assets/button/button_04.png'),
    '04-active'         : pygame.image.load('assets/button/button_04_active.png'),
    'Human'             : pygame.image.load('assets/button/button_human.png'),
    'Human-active'      : pygame.image.load('assets/button/button_human_active.png')
}
board_8 = {
    'not-numbered-light': pygame.image.load('assets/board/board_8.png'),
    'not-numbered-dark' : pygame.image.load('assets/board/board_8_dark.png'),
    'numbered-light'    : pygame.image.load('assets/board/board_8_numbered.png'),
    'numbered-dark'     : pygame.image.load('assets/board/board_8_numbered_dark.png')
}
board_10 = {
    'not-numbered-light': pygame.image.load('assets/board/board_10.png'),
    'not-numbered-dark' : pygame.image.load('assets/board/board_10_dark.png'),
    'numbered-light'    : pygame.image.load('assets/board/board_10_numbered.png'),
    'numbered-dark'     : pygame.image.load('assets/board/board_10_numbered_dark.png')
}
piece_8 = {
    'blue'              : pygame.image.load('assets/pieces/8x8_blue.png'),
    'cyan'              : pygame.image.load('assets/pieces/8x8_cyan.png'),
    'green'             : pygame.image.load('assets/pieces/8x8_green.png'),
    'orange'            : pygame.image.load('assets/pieces/8x8_orange.png'),
    'pink'              : pygame.image.load('assets/pieces/8x8_pink.png'),
    'purple'            : pygame.image.load('assets/pieces/8x8_purple.png'),
    'red'               : pygame.image.load('assets/pieces/8x8_red.png'),
    'yellow'            : pygame.image.load('assets/pieces/8x8_yellow.png')
}
piece_10 = {
    'blue'              : pygame.image.load('assets/pieces/10x10_blue.png'),
    'cyan'              : pygame.image.load('assets/pieces/10x10_cyan.png'),
    'green'             : pygame.image.load('assets/pieces/10x10_green.png'),
    'orange'            : pygame.image.load('assets/pieces/10x10_orange.png'),
    'pink'              : pygame.image.load('assets/pieces/10x10_pink.png'),
    'purple'            : pygame.image.load('assets/pieces/10x10_purple.png'),
    'red'               : pygame.image.load('assets/pieces/10x10_red.png'),
    'yellow'            : pygame.image.load('assets/pieces/10x10_yellow.png')
}
piece_hover = {
    '8x8'               : pygame.image.load('assets/pieces/8x8_hover.png'),
    '10x10'             : pygame.image.load('assets/pieces/10x10_hover.png')
}
color_picker = {
    'blue'              : pygame.image.load('assets/pieces/cp_blue.png'),
    'cyan'              : pygame.image.load('assets/pieces/cp_cyan.png'),
    'green'             : pygame.image.load('assets/pieces/cp_green.png'),
    'orange'            : pygame.image.load('assets/pieces/cp_orange.png'),
    'pink'              : pygame.image.load('assets/pieces/cp_pink.png'),
    'purple'            : pygame.image.load('assets/pieces/cp_purple.png'),
    'red'               : pygame.image.load('assets/pieces/cp_red.png'),
    'yellow'            : pygame.image.load('assets/pieces/cp_yellow.png')
}
cp_selected            =  pygame.image.load('assets/pieces/cpselected.png')


# Assets enumeration
bp_enum = list(enumerate(button_player))
p8_enum = list(enumerate(piece_8))
p10_enum = list(enumerate(piece_10))
cp_enum = list(enumerate(color_picker))


# Font definition
font_time              =  pygame.font.Font('assets/fonts/coolvetica.ttf', 100)
font_time_stack        =  pygame.font.Font('assets/fonts/coolvetica.ttf', 40)
font_player_name       =  pygame.font.Font('assets/fonts/coolvetica.ttf', 80)
font_fps               =  pygame.font.Font('assets/fonts/coolvetica.ttf', 18)
font_move_history      =  pygame.font.Font('assets/fonts/coolvetica.ttf', 30)
font_piece             =  pygame.font.Font('assets/fonts/coolvetica.ttf', 30)
font_score             =  pygame.font.Font('assets/fonts/coolvetica.ttf', 85)
font_warning           =  pygame.font.Font('assets/fonts/coolvetica.ttf', 30)


# Team AI initialization
p01     = [HalmaPlayer01A('AI 01-A'), HalmaPlayer01B('AI 01-B'), HalmaPlayer01A('AI 01-C'), HalmaPlayer01B('AI 01-D')]
p02     = [HalmaPlayer02A('AI 02-A'), HalmaPlayer02B('AI 02-B'), HalmaPlayer02A('AI 02-C'), HalmaPlayer02B('AI 02-D')]
p03     = [HalmaPlayer03A('AI 03-A'), HalmaPlayer03B('AI 03-B'), HalmaPlayer03A('AI 03-C'), HalmaPlayer03B('AI 03-D')]
p04     = [HalmaPlayer04A('AI 04-A'), HalmaPlayer04B('AI 04-B'), HalmaPlayer04A('AI 04-C'), HalmaPlayer04B('AI 04-D')]
pHuman  = [HalmaPlayerHuman('Human A'), HalmaPlayerHuman('Human B'), HalmaPlayerHuman('Human C'), HalmaPlayerHuman('Human D')]
p01[0].setTeman(p01[1])
p02[0].setTeman(p02[1])
p03[0].setTeman(p03[1])
p04[0].setTeman(p04[1])


# GUI Class
class gui:
    # GUI Assets
    screen = 0
    clock = 0
    p_8  = [0,[0],[0]]
    p_10 = [0,[0],[0]]
    dark_mode = False
    color_text = colors['TEXT']

    # Game variable
    finish = False
    winner = 0
    p1_initial = 0
    p2_initial = 0
    xi = 0
    model = 0
    running = False
    n_board = 0
    n_player = 0
    timer_time = 0
    timer_stack = 0
    move_count = 0
    move_timer = 0
    move_history = []
    move_line = [0,[],[],[],[]]
    click_start = False
    click_reset = False
    click_pause = False
    click_prev = False
    click_pmove = False
    click_human = None
    click_mode = 'piece'
    prev_pause = False
    starting = True
    shift_i = 0
    p1_selected = 0
    p2_selected = 0
    score = [0,0]
    color_click = [0,False,False]
    color_picked = [0,0,0,0,0]
    times_up = False


    # Initialization
    def __init__(self, n_board, p1, p2):
        '''
        GUI initialization with n_board size with player object array p1 and p2

        Executing halma_model initialization

        Parameter:
        - n_board : size of board
        - p1,p2 : player object array (for 2 player used first object only)
        '''

        # Player parameter initialization
        self.n_player = 2
        
        self.p1_initial = [p1[0], p1[1]]
        self.p2_initial = [p2[0], p2[1]]
        self.p1_selected = 4
        self.p2_selected = 'Human'
        self.color_picked = [0, 6, 2, 3, 0]     # Default: p1: [red,orange], p2: [green,blue]

        # Board parameter initialization
        self.n_board = n_board
        if self.n_board == 8:
            self.d = 75
            self.x0 = 623
            self.y0 = 63
        elif self.n_board == 10:
            self.d = 60
            self.x0 = 623
            self.y0 = 63
        
        # Game state
        self.starting = True
        self.running = False
        self.score = [0,0]

        # Game model
        self.model = HalmaModel()
        self.modelState = self.model.S_OK      # Running state
        self.model.awal(p1[0], p2[0])

        # Timer initialization
        self.clock = pygame.time.Clock()


    def reinit_model(self,p1,p2):
        ''' Player reinitialization with player object array p1,p2. Player number and board size can't be changed '''

        self.p1_initial = p1
        self.p2_initial = p2
        self.model = HalmaModel()
        self.modelState = self.model.S_OK      # Running state
        
        # Player selection are not same
        if self.p1_selected != self.p2_selected:
            if self.n_player == 2:
                self.model.awal(p1[0],p2[0])
            else:
                self.model.awal(p1[0],p2[0],p1[1],p2[1])
        
        # Player selection are same
        else:
            if self.n_player == 2:
                self.model.awal(p1[0],p1[2])
            else:
                self.model.awal(p1[0],p1[2],p1[1],p1[3])


    def load_template(self, numbered=None):
        ''' 
        Load game GUI template
        
        Parameter
        - Numbered (optional) (boolean): numbered board

        Included
        - Background color with colors['BG'] and ['BG_DARK']
        - Title in title variable (index 0 for light, 1 for dark)
        - Board with number or not (default numbered)
        - Start, reset, and pause button (inactive)
        - 8x8 and 10x10 button (active for selected) - Starting screen
        - Dark mode button (active for dark mode on) - Starting screen
        - Player count button (active for selected) - Starting screen
        - Player select button for p1 and p2 (active for selected) - Starting screen
            - 1-4 : AI number 1-4 (file number with suffix A/B in one team)
            - Human : Human player
        - Color picker for p1 and p2 (active for selected) - Starting screen
        - Player, score, time left, and last move info text - Playing screen
        '''

        # Background color and title
        if self.dark_mode:
            screen.fill(colors['BG_DARK'])
            screen.blit(title['dark'], dest=(35,25))
        else:
            screen.fill(colors['BG'])
            screen.blit(title['light'], dest=(35,25))
        
        # Numbered board as default
        if numbered == None:
            numbered = True

        # Board rendering
        if self.dark_mode:
            if self.n_board == 8:
                if numbered:
                    screen.blit(board_8['numbered-dark'], dest=(582,21))
                else:
                    screen.blit(board_8['not-numbered-dark'], dest=(610,50))
            elif self.n_board == 10:
                if numbered:
                    screen.blit(board_10['numbered-dark'], dest=(582,21))
                else:
                    screen.blit(board_10['not-numbered-dark'], dest=(610,50))
        else:
            if self.n_board == 8:
                if numbered:
                    screen.blit(board_8['numbered-light'], dest=(582,21))
                else:
                    screen.blit(board_8['not-numbered-light'], dest=(610,50))
            elif self.n_board == 10:
                if numbered:
                    screen.blit(board_10['numbered-light'], dest=(582,21))
                else:
                    screen.blit(board_10['not-numbered-light'], dest=(610,50))

        # Start, reset, and pause button
        screen.blit(button['start'], dest=(35,120))
        screen.blit(button['reset'], dest=(190,120))
        screen.blit(button['pause'], dest=(345,120))

        # Show on starting screen
        if self.starting:
            # 8x8 and 10x10 button
            screen.blit(button['8x8'], dest=(35,230))
            screen.blit(button['10x10'], dest=(140,230))
            
            # Dark mode button
            if self.dark_mode:
                screen.blit(button['dark-active'], dest=(35,310))
                self.color_text = colors['TEXT_DARK']
            else:
                screen.blit(button['dark'], dest=(35,310))
                self.color_text = colors['TEXT']

            # Player count button
            screen.blit(button['2p'], dest=(280,230))
            screen.blit(button['4p'], dest=(360,230))

            # AI player select button
            screen.blit(info_player['p1'], dest=(35,450))
            screen.blit(info_player['p2'], dest=(35,570))
            for i in range(0,5):
                screen.blit(button_player[bp_enum[2*i][1]], dest=(110 + i*80,445))
                screen.blit(button_player[bp_enum[2*i][1]], dest=(110 + i*80,565))

            # Color picker
            for i in range(0,8):
                screen.blit(color_picker[cp_enum[i][1]], dest=(110 + i*44,500))
                screen.blit(color_picker[cp_enum[i][1]], dest=(110 + i*44,620))
        

        # Show on playing state
        else:
            # Playing info
            screen.blit(info_text['turn'], dest=(35,200))  # Player
            screen.blit(info_text['score'], dest=(35,285))  # Score
            screen.blit(info_text['time-left'], dest=(35,365))  # Clock
            screen.blit(info_text['last-move'], dest=(35,460))  # Last Move


    def load_pieces(self, board):
        ''' Load all pieces with corresponding player color and piece id from board (2d array of board pieces coordinate) '''

        # Array traversal
        for i in range(0,self.n_board):
            for j in range(0,self.n_board):
                # Get piece number from board and parse to player_id and piece_id
                piece = board[i][j]
                player_id = piece // 100
                piece_id = piece - (player_id * 100)
                
                # Not empty block
                if piece_id != 0:
                    self.draw_piece(str(piece_id), self.color_picked[player_id], i, j)


    def draw_piece(self, text, color, i, j, opacity=None):
        ''' 
        Draw custom text and color board piece into board coordinate i,j 
        
        Parameter:
        - text (string): text to be displayed (centered with manual offset)
        - color (int): enumeration of piece color (or colors_enum-6)
            - 0(blue), 1(cyan), 2(green), 3(orange), 4(pink), 5(purple), 6(red), 7(yellow)
        - i,j: coordinate to draw (in board index i(row), j(column))
        - opacity (optional) (default=full): opacity in range 0-100 -> converted to alpha 0-255 (truncated)
        '''

        # Create text object
        textSurface = font_piece.render(text, True, (colors['WHITE']))
        textRect = textSurface.get_rect()

        # Copy board piece object
        if self.n_board == 10:
            piece = piece_10[p10_enum[color][1]].copy()
        elif self.n_board == 8:
            piece = piece_8[p8_enum[color][1]].copy()

        # Transparent
        if opacity != None:
            alpha = max(min((255 / 100) * opacity, 255),0)  # Ensure alpha value between 0-255
            piece.fill((255,255,255,alpha), None, pygame.BLEND_RGBA_MULT)

        # Draw piece to screen
        if self.n_board == 10:
            screen.blit(piece, dest=(self.x0 + j*self.d, self.y0 + i*self.d))
            textRect.center = (self.x0 + j*self.d + 27, self.y0 + i*self.d + 25)

        # Draw text to screen
        screen.blit(textSurface, textRect)

    
    def animate_move(self, initial, final):
        '''
        Animate move for smooth transition from initial to final
        '''
        pass
    

    def update_hover(self):
        ''' 
        Hover indicator for board object (piece or empty space) - currently 10x10 supported
        Blit piece_hover to hovered valid position
        
        Return: 
            - Clicked in valid position: tuple (row,column)
            - Not clicked or clicked in invalid position: None

        Mode (class variable click_mode):
        - 'piece': hover over current-player turn board-pieces 
            (references model.getBidak(row,col) and model.getGiliran())
            - left mouse button click: returning piece position (row,column)
        - 'target': hover over possible move from initial position (class variable click_human)
            (references model.bolehGeser(), model.bolehLoncat())
            - left mouse button click: returning piece position (row,column)
        '''

        # Get mouse coordinate
        mouse = pygame.mouse.get_pos()
        click = pygame.mouse.get_pressed()
        if self.click_mode == 'piece':
            if self.n_board == 10:
                for i in range(10):
                    for j in range(10):
                        if (self.in_region(mouse, (self.x0) + j*self.d, (self.y0) + i*self.d, self.d - 1, self.d - 1) 
                                            and (self.model.getBidak(i,j) // 100 == self.model.getGiliran() + 1)):
                            if click[0] and not self.click_pmove:
                                # print('[click]',i,j)
                                self.click_pmove = True
                                #self.click_human = (i,j)
                                return (i,j)
                            elif not click[0]:
                                self.click_pmove = False
                            screen.blit(piece_hover['10x10'], dest=(self.x0 + j*self.d - 4, self.y0 + i*self.d - 4))
        
        elif self.click_mode == 'target':
            if self.n_board == 10:
                for i in range(10):
                    for j in range(10):
                        if (self.in_region(mouse, (self.x0) + j*self.d, (self.y0) + i*self.d, self.d - 1, self.d - 1) 
                                            and ((i,j) in [m[-2] for m in self.model.get_all_moves(self.model.getGiliran() + 1) if m[0] == self.click_human]
                                            or self.model.getBidak(i,j) // 100 == self.model.getGiliran() + 1)):
                            if click[0] and not self.click_pmove:
                                # print('[click]',i,j)
                                self.click_pmove = True
                                return (i,j)
                            elif not click[0]:
                                self.click_pmove = False
                            screen.blit(piece_hover['10x10'], dest=(self.x0 + j*self.d - 4, self.y0 + i*self.d - 4))


    def update_possible_moves(self, p):
        '''
        Display all possible moves (intended for human playaer) with faded (60% opacity) in empty board spaces

        Parameter:
        - p: player id
        '''

        moves = []
        board = self.model.getPapan()

        # Filter duplicate final position
        for move in self.model.get_all_moves(p):
            if move[-2] not in [m[-1] for m in moves]:
                if self.click_human != None:
                    if move[0] == self.click_human:
                        moves.append(move[0:-1])
                else:
                    moves.append(move[0:-1])

        
        displayed = []
        # Display moves 
        for move in moves:
            piece_id = board[move[0][0]][move[0][1]] % 100

            for m in move:
                if m not in displayed:
                    self.draw_piece('', self.color_picked[p], m[0], m[1], 40)
                    displayed.append(m)

    
    def draw_line(self, initial, final, thick, p):
        '''
        Draw anti-aliased line from initial to final(list) with thickness thick and player id p
        
        Parameter:
        - initial (tuple): initial position
        - final (tuple): final position (can be multiple final position as 1 continuous line)
        - thick (int): thickness of line (in pixel)
        - p (int): player id (1-4)
        '''
        init = (650 + 60 * initial[1], 90 + 60 * initial[0])

        for f_idx in final:
            f = (650 + 60 * f_idx[1], 90 + 60 * f_idx[0])
            center_L = ((init[0] + f[0]) / 2, (init[1] + f[1]) / 2)
            length = math.sqrt((abs(f[0] - init[0]) ** 2) + (abs(f[1] - init[1]) ** 2))
            thickness = thick
            angle = math.atan2(init[1] - f[1], init[0] - f[0])

            UL = (center_L[0] + (length / 2.) * math.cos(angle) - (thickness / 2.) * math.sin(angle),
                center_L[1] + (thickness / 2.) * math.cos(angle) + (length / 2.) * math.sin(angle))
            UR = (center_L[0] - (length / 2.) * math.cos(angle) - (thickness / 2.) * math.sin(angle),
                center_L[1] + (thickness / 2.) * math.cos(angle) - (length / 2.) * math.sin(angle))
            BL = (center_L[0] + (length / 2.) * math.cos(angle) + (thickness / 2.) * math.sin(angle),
                center_L[1] - (thickness / 2.) * math.cos(angle) + (length / 2.) * math.sin(angle))
            BR = (center_L[0] - (length / 2.) * math.cos(angle) + (thickness / 2.) * math.sin(angle),
                center_L[1] - (thickness / 2.) * math.cos(angle) - (length / 2.) * math.sin(angle))

            pygame.gfxdraw.aapolygon(screen, (UL, UR, BR, BL), self.get_color(self.color_picked[p] + 6))
            pygame.gfxdraw.filled_polygon(screen, (UL, UR, BR, BL), self.get_color(self.color_picked[p] + 6))

            init = f


    def update_move_line(self):
        '''
        Draw move line of player pieces from class variable move_line (calls draw_line)

        move_line[p][m][pos]
        - p: player id (1-4)
        - m: move id (number of pieces have moved in last n number of moves)
        - pos: position (row,col) for move line
            - 0: initial position
            - 1-..: final position (or hop list)
        '''

        for p in range(1,5):
            for m in range(0,len(self.move_line[p])):
                init = self.move_line[p][m][0]
                final = self.move_line[p][m][1:]
                self.draw_line(init, final, 10, p)

    
    def get_color(self, i):
        ''' Get colors RGB value from colors enumeration (0 - n) '''

        return colors[colors_enum[i][1]]


    def start_animation(self):
        ''' Starting animation "PRESS START TO PLAY" with changing piece color picked by player 1 and 2 '''

        # Starting animation index
        self.xi = (self.xi + 1) % 100
        x = self.xi // 50

        # Background color
        for i in range(0,self.n_board):
            for j in range(0,self.n_board):
                self.draw_piece("",self.color_picked[2-x],i,j)
        
        # 10x10 board
        if self.n_board == 10:
            z = 0
        elif self.n_board == 8:
            z = 1

        # 'PRESS START TO PLAY' text
        self.draw_piece('P',self.color_picked[1+x], 1, 2-z)
        self.draw_piece('R',self.color_picked[1+x], 1, 3-z)
        self.draw_piece('E',self.color_picked[1+x], 1, 4-z)
        self.draw_piece('S',self.color_picked[1+x], 1, 5-z)
        self.draw_piece('S',self.color_picked[1+x], 1, 6-z)
        
        self.draw_piece('S',self.color_picked[1+x], 3, 3-z)
        self.draw_piece('T',self.color_picked[1+x], 3, 4-z)
        self.draw_piece('A',self.color_picked[1+x], 3, 5-z)
        self.draw_piece('R',self.color_picked[1+x], 3, 6-z)
        self.draw_piece('T',self.color_picked[1+x], 3, 7-z)
        
        self.draw_piece('T',self.color_picked[1+x], 5-z, 2-z)
        self.draw_piece('O',self.color_picked[1+x], 5-z, 3-z)
        
        self.draw_piece('P',self.color_picked[1+x], 7-z, 3-z)
        self.draw_piece('L',self.color_picked[1+x], 7-z, 4-z)
        self.draw_piece('A',self.color_picked[1+x], 7-z, 5-z)
        self.draw_piece('Y',self.color_picked[1+x], 7-z, 6-z)


    def reset_timer(self):
        ''' Set timer_time to initial value '''

        self.timer_time = 20000     # in millisecond
        self.clock.get_rawtime()


    def tick_timer(self):
        ''' Substract timer with elapsed time '''
        
        self.clock.tick_busy_loop(60)   # Framerate cap
        self.timer_time -= self.clock.get_rawtime()
        self.move_timer -= self.clock.get_rawtime()


    def get_timer(self):
        ''' Get timer from model calculation '''
        time_left = self.model.JATAH_WAKTU - (self.model.getJatahWaktu(self.model.getGiliran()) - self.model.getSisaWaktu())
        if time_left > 0:
            return time_left
        else:
            return 0
    

    def update_timer(self):
        ''' Update time left on screen '''

        screen.blit(font_time.render('{:.2f}'.format(self.get_timer()), True, self.color_text), dest=(210,340))
    

    def update_timer_stack(self):
        ''' Update accumulative reserved time for specific player turn on screen '''

        p = self.model.getGiliran()
        stack = self.model.getJatahWaktu(p)
        time_left = self.model.JATAH_WAKTU - (self.model.getJatahWaktu(self.model.getGiliran()) - self.model.getSisaWaktu())
        if time_left <= 0:
            stack -= (0 - time_left)
        
        if stack < 0:
            self.times_up = p + 1
            self.winner = 2 - p
            screen.blit(font_time_stack.render('(TIMES UP!)', True, self.get_color(self.color_picked[p+1] + 6)), dest=(420,355))
        else:
            screen.blit(font_time_stack.render('({:.2f})'.format(stack), True, self.get_color(self.color_picked[p+1] + 6)), dest=(420,355))


    def update_move_count(self):
        ''' Display move count (all player combined) '''
        
        screen.blit(font_fps.render('Move: {}'.format(self.move_count), True, self.color_text), dest=(1100,685))


    def update_fps(self):
        ''' Display FPS by 10 pygame timer tick average '''
        
        screen.blit(font_fps.render('FPS: {:.1f}'.format(self.clock.get_fps()), True, self.color_text), dest=(1180,685))


    def update_player(self):
        ''' Display player name with corresponding player color picked '''
        p = self.model.getGiliran()
        player = self.model.getPemain(p)
        screen.blit(font_player_name.render(player.nama, True, self.get_color(self.color_picked[p+1] + 6)), dest=(210,175))


    def update_score(self):
        ''' Player score display from class attribute score[0:2] '''

        screen.blit(font_score.render('{} - {}'.format(self.score[0], self.score[1]), True, self.color_text), dest=(210,260))


    def update_history(self):
        ''' 
        Display move history with n = 4 from attribute move_history, blit to screen
        
        move_history[hist_no][i]
            (0) turn index,
            (1) initial position,
            (2) final position (list of final position),
            (3) move action (0,1,2) - geser,loncat,berhenti,
            (4) move execution time
        '''

        n = 4   # History size
        move_hist = self.move_history.copy()
        move_hist.reverse()

        for i in range(0,len(move_hist)):
            name = self.model.getPemain(move_hist[i][0]).nama
            if move_hist[i][3] == self.model.A_GESER:
                move_type = 'GESER'
            elif move_hist[i][3] == self.model.A_LONCAT:
                move_type = 'LONCAT'
            elif move_hist[i][3] == self.model.A_BERHENTI:
                move_type = 'BERHENTI'

            # Player 1 move
            screen.blit(font_move_history.render('{} | {} {} -> {} ({:.2f}s)'.format(name, move_type, move_hist[i][1], move_hist[i][2], move_hist[i][4]),1,
                                (self.get_color(self.color_picked[move_hist[i][0] + 1] + 6))), dest=(35, 510 + i*40))


    def show_winner(self,p):
        ''' Show winner text '{team name} WIN!' after each round '''

        screen.blit(font_warning.render(self.get_object(p)[0].nama[:-2] + ' WIN!', True, self.color_text), dest=(650,670))


    def get_object(self, p):
        ''' 
        Get object with object id 
        
        Values:
            AI : 1-4 (team number)
            Human : 'Human' 
        '''

        if p == 1:
            return p01
        elif p == 2:
            return p02
        elif p == 3:
            return p03
        elif p == 4:
            return p04
        elif p == 'Human':
            return pHuman


    def in_region(self, a, x0, y0, dx, dy):       
        ''' Check position a (tuple x,y) in region box with boundary x0 and y0 (left and upper) with size of dx and dy '''

        return (x0 <= a[0] <= x0 + dx) and (y0 <= a[1] <= y0 + dy)


    def in_region_circle(self, a, x0, y0, d):
        ''' Check position a (tuple x,y) in region box with boundary x0 and y0 (left and upper) with circle diameter d '''

        cx = x0 + d/2
        cy = y0 + d/2
        dist = math.sqrt((abs(a[0] - cx)) ** 2 + (abs(a[1] - cy)) ** 2)
        return dist <= d/2


    def update_button(self):
        '''
        Button hover and click with mouse check, change to active based on state or click
        
        Include:
        - Start, reset, pause button
        - 8x8, 10x10 button
        - 2P, 4P button
        - Dark mode button
        - AI picker (1-4) and Human
        - Color picker
        '''

        # Get mouse coordinate and click
        mouse = pygame.mouse.get_pos()
        click = pygame.mouse.get_pressed()

        # Start button
        if self.in_region(mouse,35,120,126,44):
            screen.blit(button['start-active'], dest=(35,120))
            if click[0] == 1:
                # print('[click start]')
                self.action_start()

        # Reset button
        if self.in_region(mouse,190,120,126,44):
            screen.blit(button['reset-active'], dest=(190,120))
            if click[0] == 1:
                # print('[click reset]')
                self.action_reset()

        # Pause button
        if self.in_region(mouse,345,120,126,44):
            screen.blit(button['pause-active'], dest=(345,120))
            if click[0] and not self.click_prev:
                # print('[click pause]')
                self.click_prev = True
                self.action_pause()
            elif not click[0]:
                self.click_prev = False

        # Display in starting mode
        if self.starting:
            # 8x8 button
            if (self.in_region(mouse,35,230,92,50) or self.n_board == 8):
                screen.blit(button['8x8-active'], dest=(35,230))
                if click[0] == 1 and self.in_region(mouse,35,230,92,50):
                    self.n_board = 8
                    self.d = 75
                    self.x0 = 626
                    self.y0 = 66

            # 10x10 button
            if (self.in_region(mouse,140,230,104,50) or self.n_board == 10):
                screen.blit(button['10x10-active'], dest=(140,230))
                if click[0] == 1 and self.in_region(mouse,140,230,104,50):
                    self.n_board = 10
                    self.d = 60
                    self.x0 = 623
                    self.y0 = 63

            # 2P button
            if (self.in_region(mouse,280,230,67,50) or self.n_player == 2):
                screen.blit(button['2p-active'], dest=(280,230))
                if click[0] == 1 and self.in_region(mouse,280,230,67,50):
                    self.n_player = 2
            
            # 4P button
            if (self.in_region(mouse,360,230,67,50) or self.n_player == 4):
                screen.blit(button['4p-active'], dest=(360,230))
                if click[0] == 1 and self.in_region(mouse,360,230,67,50):
                    self.n_player = 4

            # Dark mode button
            if (self.in_region(mouse,35,310,208,50) or self.dark_mode == True):
                screen.blit(button['dark-active'], dest=(35,310))
                if click[0] == 1 and self.in_region(mouse,35,310,208,50):
                    self.dark_mode = not self.dark_mode
                    time.sleep(0.2)
            
            # AI picker
            for i in range(1,4):
                # Row 1 (AI)
                if self.in_region(mouse,110 + 80*i,445,69,50) or self.p1_selected == (i + 1):
                    screen.blit(button_player[bp_enum[(i * 2) + 1][1]], dest=(110 + 80*i,445))
                    if click[0] == 1 and self.in_region(mouse,110 + 80*i,445,69,50) and self.p1_selected != (i + 1):
                        # print('[{}-{} selected as p1]'.format(self.get_object(i+1)[0].nama, self.get_object(i+1)[1].nama))
                        self.p1_selected = i + 1
                
                # Row 1 (Human)
                if self.in_region(mouse,430,445,137,50) or self.p1_selected == 'Human':
                    screen.blit(button_player['Human-active'], dest=(430,445))
                    if click[0] == 1 and self.in_region(mouse,430,445,137,50) and self.p1_selected != 'Human':
                        # print('[Human is selected as p1]')
                        self.p1_selected = 'Human'

                # Row 2
                if self.in_region(mouse,110 + 80*i,565,69,50) or self.p2_selected == (i + 1):
                    screen.blit(button_player[bp_enum[(i * 2) + 1][1]], dest=(110 + 80*i,565))
                    if click[0] == 1 and self.in_region(mouse,110 + 80*i,565,69,50) and self.p2_selected != (i + 1):
                        # print('[{}-{} selected as p2]'.format(self.get_object(i+1)[0].nama, self.get_object(i+1)[1].nama))
                        self.p2_selected = i + 1

                # Row 2 (Human)
                if self.in_region(mouse,430,565,137,50) or self.p2_selected == 'Human':
                    screen.blit(button_player['Human-active'], dest=(430,565))
                    if click[0] == 1 and self.in_region(mouse,430,565,137,50) and self.p2_selected != 'Human':
                        # print('[Human is selected as p2]')
                        self.p2_selected = 'Human'

            # Color picker
            for i in range(0,8):
                # Player 1
                if self.in_region_circle(mouse,110 + 44*i,500,30) or i == self.color_picked[1] or i == self.color_picked[3]:
                    screen.blit(cp_selected, dest=(110 + 44*i,500))
                    if click[0] == 1 and self.in_region_circle(mouse,110 + 44*i,500,30):
                            self.color_picked[1] = i
                            self.color_picked[3] = i
                
                # Player 2
                if self.in_region_circle(mouse,110 + 44*i,620,30) or i == self.color_picked[2] or i == self.color_picked[4]:
                    screen.blit(cp_selected, dest=(110 + 44*i,620))
                    if click[0] == 1 and self.in_region_circle(mouse,110 + 44*i,620,30):
                        self.color_picked[2] = i
                        self.color_picked[4] = i
                

    def action_start(self):
        ''' Action when START button is clicked '''
        
        if self.n_board == 10:
            self.click_start = True
        else:
            screen.blit(font_warning.render('Currently only 10x10 board are supported', True, self.color_text), dest=(35,370))


    def action_reset(self):
        ''' Action when RESET button is clicked '''
        
        self.click_reset = True


    def action_pause(self):
        ''' Action when PAUSE button is clicked '''
        
        self.click_pause = not self.click_pause


    def update_screen(self):
        '''
        Update all screen and interaction element per frame
        
        Included:
        - Template (load_template)
        - Button checker
        - Move line
        - Board pieces
        - Hovering pieces (if player is Human)
        - Player, score, timer, timer stack and move history information
        - FPS and move count
        - Winner text (if round is finished)
        '''

        # Frame and loop timing
        self.clock.tick_busy_loop(60)   # Framerate cap

        # Check event
        self.check_event()

        self.load_template(numbered=True)
        self.update_button()
        self.update_fps()

        # Starting state only
        if self.starting:
            self.start_animation()
        
        # Playing state only
        elif self.running:
            #self.update_move_line()
            if self.p.getType() == 'Human':
                self.update_possible_moves(self.p.nomor)
                if self.click_human != None :
                    screen.blit(piece_hover['10x10'], dest=(self.x0 + self.click_human[1]*self.d - 4, 
                                                                self.y0 + self.click_human[0]*self.d - 4))
            self.load_pieces(self.model.getPapan())
            self.update_player()
            self.update_score()
            self.update_timer()
            self.update_timer_stack()
            self.update_move_count()
            self.update_history()
        
        # Round finished only
        if self.finish or self.times_up:
            self.show_winner(self.winner)
        
        pygame.display.update()
        

    def check_event(self):
        '''
        Pygame event checker (quit, mouse, keyboard, etc). 
        Terminate execution when quit event is triggered
        '''

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.running = False
                pygame.quit()
                sys.exit()


    def starting_screen(self):
        '''
        Starting screen GUI handler

        Start clicked: update running variable to True and starting variable to False
        Reset clicked: reinit model with selected player object

        - GUI screen, button, and event update
        '''

        self.update_screen()
        
        # Start button clicked
        if self.click_start:
            self.click_start = False

            # Change state
            self.starting = False
            self.running = True
            
            # Start game with AI selected (reinitialize model)
            self.reinit_model(self.get_object(self.p1_selected), self.get_object(self.p2_selected))
            self.p = self.model.getPemain(self.model.getGiliran())
        
        # Pause button clicked
        while self.click_pause:
            self.update_screen()

        # Reset button clicked (Pretty useless?)
        if self.click_reset:
            self.click_reset = False

            # Reinitialize halma model
            self.reinit_model(self.p1_initial, self.p2_initial)
    
    def get_player_move(self):
        '''
        Get player move (from AI and Human)
        
        Return: initial_pos(tuple), final_pos(tuple), action(int)
        
        Action: (model.A_GESER, model.A_LONCAT, model.A_BERHENTI)
        '''

        self.model.mainMulai()
        self.p = self.model.getPemain(self.model.getGiliran())
        
        # AI player
        if self.p.getType() == 'AI':
            self.update_screen()
            final_pos, initial_pos, action = self.p.main(self.model)

        # Human player
        elif self.p.getType() == 'Human':
            self.click_human = None
            initial_pos = None
            final = None
            click2 = None

            # Waiting for initial position click
            self.click_mode = 'piece'
            while initial_pos == None:
                self.tick_timer()
                self.update_screen()
                initial_pos = self.update_hover()
                pygame.display.update()
            self.click_human = initial_pos
            
            # Waiting for final position click
            self.click_mode = 'target'
            while final == None:
                self.tick_timer()
                self.update_screen()
                click2 = self.update_hover()
                if click2 != None:
                    if self.model.is_board_piece(click2):
                        initial_pos = click2
                        self.click_human = initial_pos
                    else:
                        final = click2
                pygame.display.update()

            # Determine type of action
            if (abs(final[0] - initial_pos[0]) ** 2 + abs(final[1] - initial_pos[1]) ** 2) > 2:
                final_pos = self.model.calc_path_hop(initial_pos, final)[1:]
                action = self.model.A_LONCAT
            elif (abs(final[0] - initial_pos[0]) ** 2 + abs(final[1] - initial_pos[1]) ** 2) > 0:
                final_pos = [final]
                action = self.model.A_GESER
            else:
                action = self.model.BERHENTI
            
        self.move_count += 1
        return initial_pos, final_pos, action

    
    def check_termination(self):
        '''
        Check if state is terminal condition (all board pieces of player in target zone)

        Reinitialize model and recursively calls main for next round
        '''

        # 2-Player
        if self.n_player == 2:
            if self.model.akhir() or self.times_up:
                pieces = self.model.getPosisiBidak(0)
                self.finish = True
                
                if self.times_up == 1:
                    # print('[P2 Menang!]')
                    self.show_winner(self.p2_selected)
                    self.score[1] += 1
                    self.winner = self.p2_selected
                elif self.times_up == 2:
                    # print('[P1 Menang!]')
                    self.show_winner(self.p1_selected)
                    self.score[0] += 1
                    self.winner = self.p1_selected
                else:
                    # Count pieces
                    k = 0
                    for piece in pieces:
                        if self.model.dalamTujuan(0,piece[0],piece[1]):
                            k += 1
                    if k == 15:
                        # print('[P1 Menang!]')
                        self.show_winner(self.p1_selected)
                        self.score[0] += 1
                        self.winner = self.p1_selected
                    else:
                        # print('[P2 Menang!]')
                        self.show_winner(self.p2_selected)
                        self.score[1] += 1
                        self.winner = self.p2_selected
                
                # Waiting for start button press
                self.times_up = False
                self.click_start = False
                while not self.click_start:
                    self.update_screen()

                # Reinitialize halma model (next round)
                modelState = self.model.S_OK
                self.reinit_model(self.get_object(self.p1_selected), self.get_object(self.p2_selected))
                self.starting = True
                self.main()

        # 4-Player
        elif self.n_player == 4:
            if self.model.akhirBeregu():
                pieces1 = self.model.getPosisiBidak(0)
                pieces3 = self.model.getPosisiBidak(2)
                self.finish = True
                
                # Count pieces in target zone (player 1 and 3)
                k = 0
                for piece in pieces1:
                    if self.model.dalamTujuan(0,piece[0],piece[1]):
                        k += 1
                for piece in pieces3:
                    if self.model.dalamTujuan(2,piece[0],piece[1]):
                        k += 1

                # If pieces count in player 1 and 3 is full
                if k == 20:
                    # print('[P1 Menang!]')
                    self.show_winner(self.p1_selected)
                    self.score[0] += 1
                    self.winner = self.p1_selected
                else:
                    # print('[P2 Menang!]')
                    self.show_winner(self.p2_selected)
                    self.score[1] += 1
                    self.winner = self.p2_selected
                
                self.update_screen()

                # Sleep for 20s
                time.sleep(20)

                # Reinitialize halma model (next round)
                modelState = self.model.S_OK
                self.reinit_model(self.get_object(self.p1_selected), self.get_object(self.p2_selected))
                self.starting = True
                self.main()


    def main(self):
        '''
        Main GUI loop, handling starting screen and playing screen
        '''

        # Starting condition
        self.reset_timer()
        modelState = self.model.S_OK
        

        # Starting state (title screen and game configuration)
        while self.starting:
            self.starting_screen()
        
        self.update_screen()

        # Playing state (board piece and game information)
        while self.running and modelState == self.model.S_OK:
            # Pause button clicked
            while self.click_pause:
                self.update_screen()
            
            # Frame and loop timing
            self.tick_timer()

            # Get player move (piece, target position, and action)
            initial_pos, final_pos, action = self.get_player_move()

            # Save move timer
            move_finish = self.model.getWaktu()
            time_exec = self.model.getJatahWaktu(self.model.getGiliran()) - self.model.getSisaWaktu()

            # Insert to move line
            if action != self.model.A_BERHENTI:
                p_num = self.p.nomor
                self.move_line[p_num].append([initial_pos] + final_pos)
                #self.move_line[p_num] = self.move_line[p_num][-8:] (limit)

            # Insert to move history (4 last move, latest on top)
            self.move_history.append([self.model.getGiliran(), initial_pos, final_pos, action, time_exec])
            if len(self.move_history) > 4:
                self.move_history = self.move_history[-4:]

            # Run move action
            if action == self.model.A_LONCAT:
                # print('[action: loncat] {} {} {}'.format(initial_pos, final_pos, action))
                for xy in final_pos:
                    modelState = self.model.mainLoncat(initial_pos[0], initial_pos[1], xy[0], xy[1])
                    initial_pos = xy
                    time.sleep(0.08)
                    self.update_screen()
            elif action == self.model.A_GESER:
                # print('[action: geser] {} {} {}'.format(initial_pos, final_pos, action))
                modelState = self.model.mainGeser(initial_pos[0], initial_pos[1], final_pos[0][0], final_pos[0][1])
            
            # Update screen frame on current state
            self.update_screen()

            # Check termination of round (all board pieces in target zone)
            self.check_termination()

            # Next turn
            modelState = self.model.ganti(move_finish)

            # Reset button clicked
            if self.click_reset:
                self.click_reset = False

                # Reinitialize halma model
                self.reinit_model(self.p1_initial,self.p2_initial)

                # Revert state (recursively call main method)
                self.starting = True
                self.running = False
                self.main()


# Execute if invoked directly from script
if __name__ == "__main__":
    # Initial value 10x10 with player1 is p03 and player 2 is p04 (changeable in starting screen)
    g = gui(n_board=10, p1=p04, p2=pHuman)
    g.main()
    pygame.quit()
